'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var protobuf = _interopDefault(require('protobufjs'));
var immutable = require('immutable');
var $protobuf = require('protobufjs/light');
var os = _interopDefault(require('os'));
var usb = _interopDefault(require('usb'));
var Debug = _interopDefault(require('debug'));
var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));
var crypto = require('crypto');
var SerialPort = _interopDefault(require('serialport'));
var DeviceLister = _interopDefault(require('nrf-device-lister'));
var MemoryMap = _interopDefault(require('nrf-intel-hex'));
var pcNrfDfuJs = require('pc-nrf-dfu-js');
var nrfjprog = _interopDefault(require('pc-nrfjprog-js'));

/*eslint-disable block-scoped-var, no-redeclare, no-control-regex, no-prototype-builtins*/

var $root = ($protobuf.roots["default"] || ($protobuf.roots["default"] = new $protobuf.Root()))
.addJSON({
  dfu: {
    nested: {
      OpCode: {
        values: {
          RESET: 0,
          INIT: 1
        }
      },
      FwType: {
        values: {
          APPLICATION: 0,
          SOFTDEVICE: 1,
          BOOTLOADER: 2,
          SOFTDEVICE_BOOTLOADER: 3
        }
      },
      HashType: {
        values: {
          NO_HASH: 0,
          CRC: 1,
          SHA128: 2,
          SHA256: 3,
          SHA512: 4
        }
      },
      Hash: {
        fields: {
          hashType: {
            rule: "required",
            type: "HashType",
            id: 1
          },
          hash: {
            rule: "required",
            type: "bytes",
            id: 2
          }
        }
      },
      InitCommand: {
        fields: {
          fwVersion: {
            type: "uint32",
            id: 1
          },
          hwVersion: {
            type: "uint32",
            id: 2
          },
          sdReq: {
            rule: "repeated",
            type: "uint32",
            id: 3
          },
          type: {
            type: "FwType",
            id: 4
          },
          sdSize: {
            type: "uint32",
            id: 5
          },
          blSize: {
            type: "uint32",
            id: 6
          },
          appSize: {
            type: "uint32",
            id: 7
          },
          hash: {
            type: "Hash",
            id: 8
          },
          isDebug: {
            type: "bool",
            id: 9,
            options: {
              "default": false
            }
          }
        }
      },
      ResetCommand: {
        fields: {
          timeout: {
            rule: "required",
            type: "uint32",
            id: 1
          }
        }
      },
      Command: {
        fields: {
          opCode: {
            type: "OpCode",
            id: 1
          },
          init: {
            type: "InitCommand",
            id: 2
          },
          reset: {
            type: "ResetCommand",
            id: 3
          }
        }
      },
      SignatureType: {
        values: {
          ECDSA_P256_SHA256: 0,
          ED25519: 1
        }
      },
      SignedCommand: {
        fields: {
          command: {
            rule: "required",
            type: "Command",
            id: 1
          },
          signatureType: {
            rule: "required",
            type: "SignatureType",
            id: 2
          },
          signature: {
            rule: "required",
            type: "bytes",
            id: 3
          }
        }
      },
      Packet: {
        fields: {
          command: {
            type: "Command",
            id: 1
          },
          signedCommand: {
            type: "SignedCommand",
            id: 2
          }
        }
      }
    }
  }
});

/* Copyright (c) 2015 - 2018, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

var root = protobuf.Root.fromJSON($root);

var InitPacket = new immutable.Record({
    fwVersion: 0xFFFFFFFF,
    hwVersion: null,
    sdReq: null,
    fwType: null,
    sdSize: 0,
    blSize: 0,
    appSize: 0,
    hashType: null,
    hash: null,
    isDebug: true,
    signatureType: null,
    signature: null,
});

var OpCode = {
    RESET: 0,
    INIT: 1,
};

var FwType = {
    APPLICATION: 0,
    SOFTDEVICE: 1,
    BOOTLOADER: 2,
    SOFTDEVICE_BOOTLOADER: 3,
};

var HashType = {
    NO_HASH: 0,
    CRC: 1,
    SHA128: 2,
    SHA256: 3,
    SHA512: 4,
};

var SignatureType = {
    ECDSA_P256_SHA256: 0,
    ED25519: 1,
};

// Create hash by using hash type and bytes
function createHash(hashType, hashInput) {
    var hashMessage = root.lookupType('dfu.Hash');
    var hashPayload = {
        hashType: hashType,
        hash: hashInput,
    };
    var hash = hashMessage.create(hashPayload);

    return hash;
}

// Create reset command
function createResetCommand(timeout) {
    var resetCommandMessage = root.lookupType('dfu.ResetCommand');
    var resetCommandPayload = {
        timeout: timeout,
    };
    var resetCommand = resetCommandMessage.create(resetCommandPayload);

    return resetCommand;
}

// Create init command
function createInitCommand(
    fwVersion,
    hwVersion,
    sdReq,
    type,
    sdSize,
    blSize,
    appSize,
    hash,
    isDebug
) {
    if ( isDebug === void 0 ) isDebug = false;

    var initCommandMessage = root.lookupType('dfu.InitCommand');
    var initCommandPayload = {
        fwVersion: fwVersion,
        hwVersion: hwVersion,
        sdReq: sdReq,
        type: type || FwType.APPLICATION,
        sdSize: sdSize,
        blSize: blSize,
        appSize: appSize,
        hash: hash,
        isDebug: isDebug,
    };

    return initCommandMessage.create(initCommandPayload);
}

// Create command
function createCommand(opCode, commandInput) {
    var commandMessage = root.lookupType('dfu.Command');
    var commandPayload = {
        opCode: opCode
    };
    commandPayload[opCode === OpCode.INIT ? 'init' : 'reset'] = commandInput;
    var command = commandMessage.create(commandPayload);

    return command;
}

// Create signed command
function createSignedCommand(command, signatureType, signature) {
    var signedCommandMessage = root.lookupType('dfu.SignedCommand');
    var signedCommandPayload = {
        command: command,
        signatureType: signatureType,
        signature: signature,
    };
    var signedCommand = signedCommandMessage.create(signedCommandPayload);

    return signedCommand;
}

// Create packet
function createPacket(command, isSigned) {
    var packetPayload = {};
    packetPayload[isSigned ? 'signedCommand' : 'command'] = command;
    var packetMessage = root.lookupType('dfu.Packet');
    var packet = packetMessage.create(packetPayload);

    return packet;
}

// Convert protocol buffer message to buffer
function messageToBuffer(type, message) {
    if (!message) {
        return undefined;
    }

    var dfuMessage = root.lookup(("dfu." + type));
    var buffer = dfuMessage.encode(message).finish();

    return buffer;
}

/**
 * Create reset command packet
 *
 * @param {Integer} timeout         the timeout for reset
 * @param {Integer} signatureType   the type of signature
 * @param {Array}   signature       the signature in bytes
 *
 * @returns {Object} reset command packet
 */
function createResetPacket(timeout, signatureType, signature) {
    try {
        // Check input parameters
        if (timeout === undefined) {
            throw new Error('Timeout is not set');
        }

        // It checks both null and undefined here
        if ((signatureType == null && signature != null) ||
            (signatureType != null && signature == null)) {
            throw new Error('Either signature type or signature is not set');
        }

        // Create reset command
        var resetCommand = createResetCommand(timeout);
        var isSigned = false;
        var command = createCommand(OpCode.RESET, resetCommand);

        // Create signed command if it is signed
        // It checks both null and undefined here
        if (signatureType != null && signature != null) {
            isSigned = true;
            command = createSignedCommand(command, signatureType, signature);
        }

        // Create packet
        var packet = createPacket(command, isSigned);

        return packet;
    } catch (err) {
        return undefined;
    }
}

/**
 * Create reset command buffer
 *
 * @param {Integer} timeout         the timeout for reset
 * @param {Integer} signatureType   the type of signature
 * @param {Array}   signature       the signature in bytes
 *
 * @returns {Buffer} converted from reset command packet
 */
function createResetPacketBuffer(timeout, signatureType, signature) {
    var packet = createResetPacket(timeout, signatureType, signature);
    var buffer = messageToBuffer('Packet', packet);

    return buffer;
}

/**
 * Create init command packet
 *
 * @param {Integer}         fwVersion       the firmware version
 * @param {Integer}         hwVersion       the hardware version
 * @param {Array}           sdReq           the softdevcie requirements
 * @param {FwType}          fwType          the firmware type
 * @param {Integer}         sdSize          the size of softDevice
 * @param {Integer}         blSize          the size of bootloader
 * @param {Integer}         appSize         the size of application
 * @param {HashType}        hashType        the type of hash
 * @param {Array}           hash            the hash in bytes
 * @param {Boolean}         isDebug         whether it is in debug mode or not
 * @param {SignatureType}   signatureType   the type of signature
 * @param {Array}           signature       the signature in bytes
 *
 * @returns {Object} init command packet
 */
function createInitPacket(
    fwVersion,
    hwVersion,
    sdReq,
    fwType,
    sdSize,
    blSize,
    appSize,
    hashType,
    hash,
    isDebug,
    signatureType,
    signature
) {
    try {
        // It checks both null and undefined here
        if ((signatureType == null && signature != null) ||
            (signatureType != null && signature == null)) {
            throw new Error('Either signature type or signature is not set');
        }

        // Create init command
        var hashInput = createHash(hashType, hash);
        var initCommand = createInitCommand(
            fwVersion,
            hwVersion,
            sdReq,
            fwType,
            sdSize,
            blSize,
            appSize,
            hashInput,
            isDebug
        );
        var command = createCommand(OpCode.INIT, initCommand);
        var isSigned = false;

        // Create signed command if it is signed
        if (signatureType != null && signature != null) {
            command = createSignedCommand(command, signatureType, signature);
            isSigned = true;
        }

        // Create packet
        var packet = createPacket(command, isSigned);

        return packet;
    } catch (err) {
        return undefined;
    }
}

/**
 * Create init command buffer
 *
 * @param {Integer}         fwVersion       the firmware version
 * @param {Integer}         hwVersion       the hardware version
 * @param {Array}           sdReq           the softdevcie requirements
 * @param {FwType}          fwType          the firmware type
 * @param {Integer}         sdSize          the size of softdevice
 * @param {Integer}         blSize          the size of bootloader
 * @param {Integer}         appSize         the size of application
 * @param {HashType}        hashType        the type of hash
 * @param {Array}           hash            the hash in bytes
 * @param {Boolean}         isDebug         whether it is in debug mode or not
 * @param {SignatureType}   signatureType   the type of signature
 * @param {Array}           signature       the signature in bytes
 *
 * @returns {Buffer} converted from init command packet
 */
function createInitPacketBuffer(
    fwVersion,
    hwVersion,
    sdReq,
    fwType,
    sdSize,
    blSize,
    appSize,
    hashType,
    hash,
    isDebug,
    signatureType,
    signature
) {
    var packet = createInitPacket(
        fwVersion,
        hwVersion,
        sdReq,
        fwType,
        sdSize,
        blSize,
        appSize,
        hashType,
        hash,
        isDebug,
        signatureType,
        signature
    );
    var buffer = messageToBuffer('Packet', packet);

    return buffer;
}

/**
 * Create init command packet Uint8Array
 *
 * @param {InitPacket} packetParams the InitPacket which carries all infomations
 *
 * @returns {Uint8Array} converted from init command packet buffer
 */
function createInitPacketUint8Array(packetParams) {
    var packet = createInitPacket(
        packetParams.fwVersion,
        packetParams.hwVersion,
        packetParams.sdReq,
        packetParams.fwType,
        packetParams.sdSize,
        packetParams.blSize,
        packetParams.appSize,
        packetParams.hashType,
        packetParams.hash,
        packetParams.isDebug,
        packetParams.signatureType,
        packetParams.signature
    );
    var buffer = messageToBuffer('Packet', packet);

    return new Uint8Array(buffer);
}

var initPacket = /*#__PURE__*/Object.freeze({
  InitPacket: InitPacket,
  OpCode: OpCode,
  FwType: FwType,
  HashType: HashType,
  SignatureType: SignatureType,
  createResetPacket: createResetPacket,
  createResetPacketBuffer: createResetPacketBuffer,
  createInitPacket: createInitPacket,
  createInitPacketBuffer: createInitPacketBuffer,
  createInitPacketUint8Array: createInitPacketUint8Array
});

/* Copyright (c) 2010 - 2018, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

var debug = Debug('device-setup:trigger');

var ReqTypeInterfaceClass = usb.LIBUSB_REQUEST_TYPE_CLASS | usb.LIBUSB_RECIPIENT_INTERFACE;
var ReqTypeIN = ReqTypeInterfaceClass | usb.LIBUSB_ENDPOINT_IN;
var ReqTypeOUT = ReqTypeInterfaceClass | usb.LIBUSB_ENDPOINT_OUT;

var NORDIC_SEM_VER_REQUEST = 8;
var NORDIC_DFU_INFO_REQUEST = 7;
var DFU_DETACH_REQUEST = 0;

var nordicInfoStructSize = 24; // 5 DWORD and 2 WORD
var detachReqBuf = Buffer.from('0');

/*
 * Returns a multibyte value from an array of bytes
 * getBytes( [1,2,3,4,5,6], 3, 2 ) => 4 + 5<<8 => 36
 */
function getBytes(array, index, length) {
    return array
        .slice(index, index + length)
        .reduce(function (c, v, i) { return c + (v << (i * 8)); }, 0);
}

var openDecorator = function (decoratee) { return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    var usbdev = args[0];
    if (usbdev.interfaces instanceof Array) {
        return decoratee.apply(void 0, args);
    }
    try {
        usbdev.open();
    } catch (e) { debug(e.message); }
    return decoratee.apply(void 0, args)
        .then(function (result) {
            try {
                usbdev.close();
            } catch (e) { debug(e.message); }
            return result;
        });
}; };

/**
 * Asserts that the given interface number exists in the given (open) USB Device,
 * and that the class, subclass and protocol match.
 * Returns nothing on success, but throws an error if the assertion fails.
 *
 * @param {Device} usbdev Instance of USB's Device
 * @param {number} interfaceNumber 0-indexed interface number to check
 * @returns {undefined}
 */
function assertDfuTriggerInterface(usbdev, interfaceNumber) {
    if (!(usbdev.interfaces instanceof Array)) {
        throw new Error('USB Device must be open before performing any operations on the DFU trigger interface');
    }
    var iface = usbdev.interfaces[interfaceNumber];
    if (!iface) {
        throw new Error(("Interface number " + interfaceNumber + " does not exist on USB device; cannot perform DFU trigger operation."));
    }

    if (iface.descriptor.bInterfaceClass !== 255 ||
        iface.descriptor.bInterfaceSubClass !== 1 ||
        iface.descriptor.bInterfaceProtocol !== 1
    ) {
        throw new Error(("Interface number " + interfaceNumber + " does not look like a DFU trigger interface; cannot perform DFU trigger operation."));
    }
}

function getDFUInterfaceNumber(usbdev) {
    var wasClosed = !(usbdev.interfaces instanceof Array);
    if (wasClosed) {
        try {
            usbdev.open();
        } catch (error) {
            debug(error.message);
            return -1;
        }
    }

    var dfuTriggerInterface = usbdev.interfaces.findIndex(function (iface) { return (
        iface.descriptor.bInterfaceClass === 255 &&
        iface.descriptor.bInterfaceSubClass === 1 &&
        iface.descriptor.bInterfaceProtocol === 1
    ); });

    if (wasClosed) {
        try {
            usbdev.close();
        } catch (error) {
            debug(error.message);
        }
    }

    return dfuTriggerInterface;
}

var getSemVersion = openDecorator(function (usbdev, interfaceNumber) { return (
    new Promise(function (resolve, reject) {
        assertDfuTriggerInterface(usbdev, interfaceNumber);
        usbdev.controlTransfer(
            ReqTypeIN,
            NORDIC_SEM_VER_REQUEST, 0, interfaceNumber, 256, function (error, data) { return (
                error
                    ? reject(error)
                    : resolve(String.fromCharCode.apply('utf16le', data).replace(/\0$/, ''))
            ); }
        );
    })
); });

var getDfuInfo = openDecorator(function (usbdev, interfaceNumber) { return (
    new Promise(function (resolve, reject) {
        assertDfuTriggerInterface(usbdev, interfaceNumber);
        usbdev.controlTransfer(
            ReqTypeIN,
            NORDIC_DFU_INFO_REQUEST, 0, interfaceNumber, nordicInfoStructSize,
            function (error, data) { return (
                error
                    ? reject(error)
                    : resolve({
                        address: getBytes(data, 0, 4),
                        firmwareSize: getBytes(data, 4, 4),
                        versionMajor: getBytes(data, 8, 2),
                        versionMinor: getBytes(data, 10, 2),
                        firmwareID: getBytes(data, 12, 4),
                        flashSize: getBytes(data, 16, 4),
                        flashPageSize: getBytes(data, 20, 4),
                    })
            ); }
        );
    })
); });

var sendDetachRequest = openDecorator(function (usbdev, interfaceNumber) { return (
    new Promise(function (resolve, reject) {
        assertDfuTriggerInterface(usbdev, interfaceNumber);
        debug(("Claiming interface " + interfaceNumber));
        usbdev.interfaces[interfaceNumber].claim();
        debug('Sending DFU detach request ');
        usbdev.controlTransfer(
            ReqTypeOUT, DFU_DETACH_REQUEST, 0, interfaceNumber, detachReqBuf,
            function (err) {
                debug(("Releasing interface " + interfaceNumber));
                usbdev.interfaces[interfaceNumber].release(function (err2) {
                    if (err2) { reject(err2); }

                    // On Windows, if the detach is successful,
                    // the target device will reboot before sending a response,
                    // so the expected result is that the control transfer will stall.
                    // On MacOS, DFU detach request does not stall as on Windows.
                    // Just regard it as detaching successfully.
                    if (err &&
                        err.errno === usb.LIBUSB_TRANSFER_STALL &&
                        err.message === 'LIBUSB_TRANSFER_STALL') {
                        resolve();
                    } else if (err &&
                        err.errno === usb.LIBUSB_ERROR_IO &&
                        err.message === 'LIBUSB_ERROR_IO') {
                        // This edge case only happens when using the "libusb" kernel
                        // driver on win32 (not "winusb", not "libusbk")
                        resolve();
                    } else if (os.platform() === 'darwin') {
                        resolve();
                    } else {
                        debug('DFU detach request did not stall as expected');
                        reject(new Error('USB DFU detach request sent, but device does not seem to have rebooted'));
                    }
                });
            }
        );
    })
); });

/**
 * Sends a detach request to a nRF USB device running in application mode,
 * and waits until it reboots out of application mode.
 *
 * @param {Device} usbdev Instance of USB's Device
 * @param {number} timeout Timeout, in milliseconds, to wait for device detachment
 * @return {Promise} Resolves to undefined
 */
function detach(usbdev, timeout) {
    if ( timeout === void 0 ) timeout = 5000;

    debug('detach', timeout);
    return new Promise(function (resolve, reject) {
        var timeoutId;
        function checkDetachment(detachedDev) {
            if (usbdev === detachedDev) {
                debug('Detachment successful');
                clearTimeout(timeoutId);
                usb.removeListener('detach', checkDetachment);
                resolve();
            }
        }

        timeoutId = setTimeout(function () {
            debug('Timeout when waiting for USB detach event');
            usb.removeListener('detach', checkDetachment);
            reject(new Error('USB detach request sent, timeout while waiting for device reboot'));
        }, timeout);

        usb.on('detach', checkDetachment);
        var dfuIface = getDFUInterfaceNumber(usbdev);

        sendDetachRequest(usbdev, dfuIface).catch(function (err) {
            debug('Error when sending detach request');
            clearTimeout(timeoutId);
            usb.removeListener('detach', checkDetachment);
            reject(err);
        });
    });
}

var dfuTrigger = /*#__PURE__*/Object.freeze({
  detach: detach,
  getDFUInterfaceNumber: getDFUInterfaceNumber,
  getSemVersion: getSemVersion,
  getDfuInfo: getDfuInfo,
  sendDetachRequest: sendDetachRequest
});

function __async(g){return new Promise(function(s,j){function c(a,x){try{var r=g[x?"throw":"next"](a);}catch(e){j(e);return}r.done?s(r.value):Promise.resolve(r.value).then(c,d);}function d(e){c(e,1);}c();})}

var debug$1 = Debug('device-setup:jprog');

var DeviceFamily = {};
DeviceFamily[nrfjprog.NRF51_FAMILY] = 'nrf51';
DeviceFamily[nrfjprog.NRF52_FAMILY] = 'nrf52';
DeviceFamily[nrfjprog.NRF91_FAMILY] = 'nrf91';

var DeviceType = {};
DeviceType[nrfjprog.NRF51xxx_xxAA_REV1] = 'NRF51xxx_xxAA_REV1';
DeviceType[nrfjprog.NRF51xxx_xxAA_REV2] = 'NRF51xxx_xxAA_REV2';
DeviceType[nrfjprog.NRF51xxx_xxAA_REV3] = 'NRF51xxx_xxAA_REV3';
DeviceType[nrfjprog.NRF51xxx_xxAB_REV3] = 'NRF51xxx_xxAB_REV3';
DeviceType[nrfjprog.NRF51xxx_xxAC_REV3] = 'NRF51xxx_xxAC_REV3';
DeviceType[nrfjprog.NRF9160_xxAA_FUTURE] = 'NRF9160_xxAA_FUTURE';
DeviceType[nrfjprog.NRF9160_xxAA_REV1] = 'NRF9160_xxAA_REV1';
DeviceType[nrfjprog.NRF51801_xxAB_REV3] = 'NRF51801_xxAB_REV3';
DeviceType[nrfjprog.NRF51802_xxAA_REV3] = 'NRF51802_xxAA_REV3';
DeviceType[nrfjprog.NRF52810_xxAA_FUTURE] = 'NRF52810_xxAA_FUTURE';
DeviceType[nrfjprog.NRF52810_xxAA_REV1] = 'NRF52810_xxAA_REV1';
DeviceType[nrfjprog.NRF52810_xxAA_REV2] = 'NRF52810_xxAA_REV2';
DeviceType[nrfjprog.NRF52811_xxAA_FUTURE] = 'NRF52811_xxAA_FUTURE';
DeviceType[nrfjprog.NRF52811_xxAA_REV1] = 'NRF52811_xxAA_REV1';
DeviceType[nrfjprog.NRF52832_xxAA_ENGA] = 'NRF52832_xxAA_ENGA';
DeviceType[nrfjprog.NRF52832_xxAA_ENGB] = 'NRF52832_xxAA_ENGB';
DeviceType[nrfjprog.NRF52832_xxAA_FUTURE] = 'NRF52832_xxAA_FUTURE';
DeviceType[nrfjprog.NRF52832_xxAA_REV1] = 'NRF52832_xxAA_REV1';
DeviceType[nrfjprog.NRF52832_xxAA_REV2] = 'NRF52832_xxAA_REV2';
DeviceType[nrfjprog.NRF52832_xxAB_FUTURE] = 'NRF52832_xxAB_FUTURE';
DeviceType[nrfjprog.NRF52832_xxAB_REV1] = 'NRF52832_xxAB_REV1';
DeviceType[nrfjprog.NRF52832_xxAB_REV2] = 'NRF52832_xxAB_REV2';
DeviceType[nrfjprog.NRF52840_xxAA_ENGA] = 'NRF52840_xxAA_ENGA';
DeviceType[nrfjprog.NRF52840_xxAA_ENGB] = 'NRF52840_xxAA_ENGB';
DeviceType[nrfjprog.NRF52840_xxAA_FUTURE] = 'NRF52840_xxAA_FUTURE';
DeviceType[nrfjprog.NRF52840_xxAA_REV1] = 'NRF52840_xxAA_REV1';
DeviceType[nrfjprog.NRF52840_xxAA_REV2] = 'NRF52840_xxAA_REV2';

function parseSerial(serialNumber) {
    return parseInt(serialNumber, 10);
}

function read(serialNumber, address, length) {
    return new Promise(function (resolve, reject) {
        nrfjprog.read(parseSerial(serialNumber), address, length, function (err, contents) {
            if (err) {
                reject(err);
            } else {
                resolve(contents);
            }
        });
    });
}

function getDeviceInfo(serialNumber) {
    return new Promise(function (resolve, reject) {
        nrfjprog.getDeviceInfo(parseSerial(serialNumber), function (err, deviceInfo) {
            if (err) {
                reject(new Error(("Error when getting device info " + (err.message))));
            } else {
                var family = DeviceFamily[deviceInfo.family];
                var deviceType = DeviceType[deviceInfo.deviceType];
                if (!family) {
                    reject(new Error('Couldn\'t get device family'));
                }
                resolve(Object.assign({}, deviceInfo, { family: family, deviceType: deviceType }));
            }
        });
    });
}

/**
 * Program the device with the given serial number with the given firmware
 * using nrfjprog.
 *
 * @param {String|Number} serialNumber The serial number of the device.
 * @param {String|Buffer} firmware Firmware path or firmware contents as buffer.
 * @returns {Promise} Promise that resolves if successful or rejects with error.
 */
function program(serialNumber, firmware) {
    var fw;
    var options = {};
    if (firmware instanceof Buffer) {
        fw = firmware.toString('utf-8');
        var INPUT_FORMAT_HEX_STRING = 1;
        options.inputFormat = INPUT_FORMAT_HEX_STRING;
    } else {
        fw = firmware;
    }
    return new Promise(function (resolve, reject) {
        nrfjprog.program(parseSerial(serialNumber), fw, options, function (err) {
            if (err) {
                reject(err);
            } else {
                resolve();
            }
        });
    });
}

/**
 * Try to open and close the given serial port to see if it is available. This
 * is needed to identify if a SEGGER J-Link device is in a bad state. If
 * pc-nrfjprog-js tries to interact with a device in bad state, it will hang
 * indefinitely.
 *
 * @param {object} device Device object, ref. nrf-device-lister.
 * @returns {Promise} Promise that resolves if available, and rejects if not.
 */
function verifySerialPortAvailable(device) {
    if (!device.serialport) {
        return Promise.reject(new Error('No serial port available for device with '
            + "serial number " + (device.serialNumber)));
    }
    return new Promise(function (resolve, reject) {
        var serialPort = new SerialPort(device.serialport.comName, { autoOpen: false });
        serialPort.open(function (openErr) {
            if (openErr) {
                reject(openErr);
            } else {
                serialPort.close(function (closeErr) {
                    if (closeErr) {
                        reject(closeErr);
                    } else {
                        resolve();
                    }
                });
            }
        });
    });
}

function openJLink(device) {
    return new Promise(function (resolve, reject) {
        nrfjprog.open(parseSerial(device.serialNumber), function (err) { return (err ? reject(err) : resolve()); });
    });
}

function closeJLink(device) {
    return new Promise(function (resolve, reject) {
        nrfjprog.close(parseSerial(device.serialNumber), function (err) { return (err ? reject(err) : resolve()); });
    });
}

function validateFirmware(device, firmwareFamily) {return __async(function*(){
    var fwIdAddress = firmwareFamily.fwIdAddress;
    var fwVersion = firmwareFamily.fwVersion;
    var contents;

    try {
        contents = yield read(device.serialNumber, fwIdAddress, fwVersion.length);
    } catch (error) {
        throw new Error(("Error when validating firmware " + (error.message)));
    }

    if (typeof fwVersion === 'object' && typeof fwVersion.validator === 'function') {
        return fwVersion.validator(contents);
    }

    var data = Buffer.from(contents).toString();
    return (data === fwVersion);
}())}

function programFirmware(device, firmwareFamily) {return __async(function*(){
    try {
        debug$1(("Programming " + (device.serialNumber) + " with " + (firmwareFamily.fw)));
        yield program(device.serialNumber, firmwareFamily.fw);
    } catch (programError) {
        throw new Error(("Error when programming " + (programError.message)));
    }
    return device;
}())}

/**
 * @const {number} DEFAULT_DEVICE_WAIT_TIME Default wait time for UART port to
 * show up in operating system
 */
var DEFAULT_DEVICE_WAIT_TIME = 10000;
var BASEPATH = path.dirname(require.resolve('.'));
var LATEST_BOOTLOADER = 'graviton_bootloader_v1.0.1-[nRF5_SDK_15.0.1-1.alpha_f76d012].zip';
var LATEST_BOOTLOADER_PATH = path.resolve(BASEPATH, '../bootloader', LATEST_BOOTLOADER);
var LATEST_BOOTLOADER_VERSION = 3; // check with nrfutil pkg display ...

var getDFUInterfaceNumber$1 = getDFUInterfaceNumber;
var getSemVersion$1 = getSemVersion;
var detach$1 = detach;

var InitPacket$1 = InitPacket;
var FwType$1 = FwType;
var HashType$1 = HashType;
var createInitPacketUint8Array$1 = createInitPacketUint8Array;

var openJLink$1 = openJLink;
var closeJLink$1 = closeJLink;
var verifySerialPortAvailable$1 = verifySerialPortAvailable;
var getDeviceInfo$1 = getDeviceInfo;
var validateFirmware$1 = validateFirmware;
var programFirmware$1 = programFirmware;

var debug$2 = Debug('device-setup');
var debugError = Debug('device-setup:error');

/**
 * Aux function. Returns a promise that resolves after the given time.
 *
 * @param {number} ms Time, in milliseconds, to wait until promise resolution
 * @returns {Promise<undefined>} Promise that resolves after a time
 */
function sleep(ms) {
    return new Promise(function (resolve) { return setTimeout(resolve, ms); });
}

/**
 * Check if the device is currently running DFU Bootloader
 *
 * @param {object} device nrf-device-lister device
 * @returns {boolean} true if device is currently in DFU Bootloader
 */
function isDeviceInDFUBootloader(device) {
    if (!device) {
        return false;
    }
    if (device.usb) {
        var ref = device.usb.device;
        var d = ref.deviceDescriptor;
        return (d.idVendor === 0x1915 && d.idProduct === 0x521f);
    }
    if (device.serialport) {
        var ref$1 = device.serialport;
        var vendorId = ref$1.vendorId;
        var productId = ref$1.productId;
        return (vendorId === '1915' && productId.toUpperCase() === '521F');
    }
    return false;
}

/**
 * Waits until a device (with a matching serial number) is listed by
 * nrf-device-lister, up to a maximum of `timeout` milliseconds.
 *
 * If `expectedTraits` is given, then the device must (in addition to
 * a matching serial number) also have the given traits. See the
 * nrf-device-lister library for the full list of traits.
 *
 * @param {string} serialNumber of the device expected to appear
 * @param {number} [timeout] Timeout, in milliseconds, to wait for device enumeration
 * @param {Array} [expectedTraits] The traits that the device is expected to have
 * @returns {Promise} resolved to the expected device
 */
function waitForDevice(serialNumber, timeout, expectedTraits) {
    if ( timeout === void 0 ) timeout = DEFAULT_DEVICE_WAIT_TIME;
    if ( expectedTraits === void 0 ) expectedTraits = ['serialport'];

    debug$2(("Will wait for device " + serialNumber));

    return new Promise(function (resolve, reject) {
        var timeoutId;
        var lister = new DeviceLister({
            nordicUsb: true, nordicDfu: true, serialport: true,
        });

        function checkConflation(deviceMap) {
            var device = deviceMap.get(serialNumber);
            if (device && expectedTraits.every(function (trait) { return device.traits.includes(trait); })) {
                clearTimeout(timeoutId);
                lister.removeListener('conflated', checkConflation);
                lister.removeListener('error', debugError);
                lister.stop();
                debug$2(("... found " + serialNumber));
                resolve(device);
            }
        }

        timeoutId = setTimeout(function () {
            debug$2(("Timeout when waiting for attachment of device with serial number " + serialNumber));
            lister.removeListener('conflated', checkConflation);
            lister.removeListener('error', debugError);
            lister.stop();
            reject(new Error(("Timeout while waiting for device  " + serialNumber + " to be attached and enumerated")));
        }, timeout);

        lister.on('error', debugError);
        lister.on('conflated', checkConflation);
        lister.start();
    });
}

/**
 * Sends a detach request to a device and waits until it gets reattached.
 *
 * @param {object} usbdev instance of usb device
 * @param {number} interfaceNumber of the trigger interface
 * @param {string} serialNumber of the device expected after reattach
 * @return {Promise} resolves to reattached device
 */
function detachAndWaitFor(usbdev, interfaceNumber, serialNumber) {
    debug$2('Sending detach, will wait for attach');
    return detach$1(usbdev)
        .then(function () { return waitForDevice(serialNumber); });
}

/**
 * Calculates SHA256 hash of image
 *
 * @param {Uint8Array} image to calculate hash from
 * @return {Buffer} SHA256 hash
 */
function calculateSHA256Hash(image) {
    var digest = crypto.createHash('sha256');
    digest.update(image);
    return Buffer.from(digest.digest().reverse());
}


/**
 * Loads firmware image from HEX file
 *
 * @param {Buffer|string} firmware contents of HEX file if Buffer otherwhise path of HEX file
 * @return {Uint8Array} the loaded firmware
 */
function parseFirmwareImage(firmware) {
    var contents = (firmware instanceof Buffer) ? firmware : fs.readFileSync(firmware);
    var memMap = MemoryMap.fromHex(contents);
    var startAddress;
    var endAddress;
    memMap.forEach(function (block, address) {
        startAddress = !startAddress ? address : startAddress;
        endAddress = address + block.length;
    });
    return memMap.slicePad(startAddress, Math.ceil((endAddress - startAddress) / 4) * 4);
}

/**
 * Ensures that device has a serialport that is ready to be opened
 * @param {object} device nrf-device-lister's device
 * @param {boolean} needSerialport indicates if the device is expected to have a serialport
 * @returns {Promise} resolved to device
 */
function validateSerialPort(device, needSerialport) {return __async(function*(){
    if (!needSerialport) {
        debug$2('device doesn`t need serialport');
        return device;
    }

    var checkOpen = function (comName) { return new Promise(function (resolve) {
        var port = new SerialPort(comName, { baudRate: 115200 }, function (err) {
            if (!err) { port.close(); }
            resolve(!err);
        });
    }); };

    for (var i = 10; i > 1; i -= 1) {
        /* eslint-disable-next-line no-await-in-loop */
        yield sleep(2000 / i);
        debug$2('validating serialport', device.serialport.comName, i);
        /* eslint-disable-next-line no-await-in-loop */
        if (yield checkOpen(device.serialport.comName)) {
            debug$2('resolving', device);
            return device;
        }
    }
    throw new Error('couldn`t open serialport');
}())}

/**
 * Prepares a device which is expected to be in DFU Bootlader.
 * First it loads the firmware from HEX file specified by dfu argument,
 * then performs the DFU operation.
 * This causes the device to be detached, so finally it waits for it to be attached again.
 *
 * @param {object} device nrf-device-lister's device
 * @param {object} dfu configuration object for performing the DFU
 * @returns {Promise} resolved to prepared device
 */
function prepareInDFUBootloader(device, dfu) {return __async(function*(){
    var ref = device.serialport;
    var comName = ref.comName;
    debug$2(((device.serialNumber) + " on " + comName + " is now in DFU-Bootloader..."));

    var application = dfu.application;
    var softdevice = dfu.softdevice;
    var params = dfu.params;
    params = params || {};

    var firmwareUpdates = [];
    if (softdevice) {
        var firmwareImage$1 = parseFirmwareImage(softdevice);

        var initPacketParams$1 = new InitPacket$1()
            .set('fwType', FwType$1.SOFTDEVICE)
            .set('fwVersion', 0xffffffff)
            .set('hwVersion', params.hwVersion || 52)
            .set('hashType', HashType$1.SHA256)
            .set('hash', calculateSHA256Hash(firmwareImage$1))
            .set('sdSize', firmwareImage$1.length)
            .set('sdReq', params.sdReq || []);

        var packet$1 = createInitPacketUint8Array$1(initPacketParams$1);
        firmwareUpdates.push({ initPacket: packet$1, firmwareImage: firmwareImage$1 });
    }

    var firmwareImage = parseFirmwareImage(application);

    var initPacketParams = new InitPacket$1()
        .set('fwType', FwType$1.APPLICATION)
        .set('fwVersion', params.fwVersion || 4)
        .set('hwVersion', params.hwVersion || 52)
        .set('hashType', HashType$1.SHA256)
        .set('hash', calculateSHA256Hash(firmwareImage))
        .set('appSize', firmwareImage.length)
        .set('sdReq', params.sdId || []);

    var packet = createInitPacketUint8Array$1(initPacketParams);
    firmwareUpdates.push({ initPacket: packet, firmwareImage: firmwareImage });

    var usbSerialTransport = new pcNrfDfuJs.DfuTransportUsbSerial(device.serialNumber, 0);
    var dfuOperation = new pcNrfDfuJs.DfuOperation(new pcNrfDfuJs.DfuUpdates(firmwareUpdates), usbSerialTransport);

    debug$2('Starting DFU');
    yield dfuOperation.start(true);
    debug$2('DFU completed successfully!');

    return waitForDevice(device.serialNumber, DEFAULT_DEVICE_WAIT_TIME, ['serialport', 'nordicUsb']);
}())}

/**
 * Helper function that calls optional user defined confirmation e.g. dialog or inquirer.
 *
 * @param {function} promiseConfirm Promise returning function
 * @returns {Promise} resolves to boolean
 */
function confirmHelper(promiseConfirm) {return __async(function*(){
    if (!promiseConfirm) { return true; }
    try {
        return yield promiseConfirm('Device must be programmed, do you want to proceed?');
    } catch (err) {
        throw new Error('Preparation cancelled by user');
    }
}())}

/**
 * Helper function that calls optional user defined choice e.g. dialog or inquirer.
 *
 * @param {array} choices array of choices
 * @param {function} promiseChoice Promise returning function
 * @returns {Promise} resolves to user selected choice or first element
 */
function choiceHelper(choices, promiseChoice) {return __async(function*(){
    if (choices.length > 1 && promiseChoice) {
        return promiseChoice('Which firmware do you want to program?', choices);
    }
    return choices.pop();
}())}

/**
 * Get firmware version of currently running dfu bootloader.
 *
 * @param {Object} device device
 * @returns {Promise<number>} version number of bootloader
 */
function getBootloaderVersion(device) {return __async(function*(){
    var usbSerialTransport = new pcNrfDfuJs.DfuTransportUsbSerial(device.serialNumber, 0);
    var firmwareVersions = yield usbSerialTransport.getAllFirmwareVersions();
    yield new Promise(function (resolve) { return usbSerialTransport.port.close(resolve); });

    var bootloaderVersion = firmwareVersions.find(function (fw) { return fw.imageType === 'Bootloader'; });
    if (!bootloaderVersion) {
        throw new Error('Bootloader version couldn`t be found');
    }

    debug$2(("Bootloader version " + (bootloaderVersion.version) + " is found."));
    return bootloaderVersion.version;
}())}

/**
 * Updates the dfu bootloader.
 * The bootloader is loaded from signed official zip file contained in this module.
 *
 * @param {Object} device device
 * @returns {Promise<Object>} device object after dfu is completed and device is enumerated again.
 */
function updateBootloader(device) {return __async(function*(){
    var ref = device.serialport;
    var comName = ref.comName;
    debug$2(("Bootloader for device " + (device.serialNumber) + " on " + comName + " will be updated"));

    var updates = yield pcNrfDfuJs.DfuUpdates.fromZipFilePath(LATEST_BOOTLOADER_PATH);
    var usbSerialTransport = new pcNrfDfuJs.DfuTransportUsbSerial(device.serialNumber, 0);
    var dfuOperation = new pcNrfDfuJs.DfuOperation(updates, usbSerialTransport);

    debug$2('Starting Bootloader DFU');
    yield dfuOperation.start(true);
    debug$2('Bootloader DFU completed successfully!');

    return waitForDevice(device.serialNumber, DEFAULT_DEVICE_WAIT_TIME, ['serialport']);
}())}

/**
 * Procedure of checking firmware version of the currently running bootloader,
 * in case it's not the latest - after confirmation - it is updated.
 *
 * @param {Object} device device
 * @param {function} promiseConfirm funtion that returns Promise<boolean> for confirmation
 * @returns {Promise<Object>} updated device
 */
function checkConfirmUpdateBootloader(device, promiseConfirm) {return __async(function*(){
    if (!promiseConfirm) {
        // without explicit consent bootloader will not be updated
        return device;
    }
    var bootloaderVersion = yield getBootloaderVersion(device);
    if (bootloaderVersion >= LATEST_BOOTLOADER_VERSION) {
        return device;
    }
    if (!(yield promiseConfirm('Newer version of the bootloader is available, do you want to update it?'))) {
        debug$2('Continuing with old bootloader');
        return device;
    }
    return updateBootloader(device);
}())}

/**
 * Trigger DFU Bootloader mode if the device is not yet in that mode.
 *
 * @param {Object} device device
 * @returns {Promise<Object>} device object which is already in bootloader.
 */
function ensureBootloaderMode(device) {return __async(function*(){
    var serialNumber = device.serialNumber;
    if (isDeviceInDFUBootloader(device)) {
        debug$2('Device is in bootloader mode');
        return device;
    }
    var usbdev = device.usb;
    var retry = 0;
    while (!usbdev && retry < 3) {
        retry += 1;
        debug$2('missing usb, looking for it again');
        /* eslint-disable-next-line no-await-in-loop */
        usbdev = yield waitForDevice(serialNumber, DEFAULT_DEVICE_WAIT_TIME, ['nordicUsb']).usb;
    }
    if (!usbdev) {
        throw new Error('Couldn`t recognize expected nordic usb device');
    }
    debug$2('Trying to trigger bootloader mode');
    return detachAndWaitFor(
        usbdev.device,
        getDFUInterfaceNumber$1(device.usb.device),
        serialNumber
    );
}())}

/**
 * Adds detailed output if enabled in options
 *
 * @param {Object} device device
 * @param {Object} details device
 * @param {boolean} detailedOutput device
 * @returns {Object} Either the device or the {device, details} object
 */
var createReturnValue = function (device, details, detailedOutput) { return (
    detailedOutput ? { device: device, details: details } : device
); };

/**
 * DFU procedure which also tries to update bootloader in case bootloader mode is
 * set during the process and it happens to be outdated.
 *
 * @param {Object} selectedDevice device
 * @param {Object} options options
 * @returns {Promise} device or { device, details } object
 */
function performDFU(selectedDevice, options) {return __async(function*(){
    var dfu = options.dfu;
    var needSerialport = options.needSerialport;
    var detailedOutput = options.detailedOutput;
    var promiseConfirm = options.promiseConfirm;
    var promiseConfirmBootloader = options.promiseConfirmBootloader;
    var promiseChoice = options.promiseChoice;
    var isConfirmed = yield confirmHelper(promiseConfirm);
    if (!isConfirmed) {
        // go on without DFU
        return createReturnValue(selectedDevice, { wasProgrammed: false }, detailedOutput);
    }
    var choice = yield choiceHelper(Object.keys(dfu), promiseChoice);

    try {
        var device = yield ensureBootloaderMode(selectedDevice);
        device = yield checkConfirmUpdateBootloader(
            device,
            promiseConfirmBootloader || promiseConfirm
        );
        device = yield ensureBootloaderMode(device);
        device = yield prepareInDFUBootloader(device, dfu[choice]);
        device = yield validateSerialPort(device, needSerialport);

        debug$2('DFU finished: ', device);
        return createReturnValue(device, { wasProgrammed: true }, detailedOutput);
    } catch (err) {
        debug$2('DFU failed: ', err);
        throw err;
    }
}())}

/**
 * Prepares a device listed by nrf-device-lister with expected application firmware
 * configured by options for different device types.
 * Based on the device type it decides whether it should be programmed by DFU or JProg.
 * Resolves if the setup was successful. Rejects with an error if the setup failed.
 * If detailedOutput is enabled, it will resolve with an object on the form
 * { device: {}, details: { wasProgrammed: [true|false] } }. If not, it will resolve
 * with the device object only.
 *
 * @example
 * const preparedDevice = await setupDevice(selectedDevice,
 *     {
 *         dfu: {
 *             // can have several firmwares defined, the key will be offered to choose from
 *             pca10059: {
 *                 fw: path.resolve(__dirname, 'fw/rssi-10059.hex'),
 *                 semver: 'rssi_cdc_acm 2.0.0+dfuMar-27-2018-12-41-04',
 *             },
 *         },
 *         jprog: {
 *             nrf52: {
 *                 fw: path.resolve(__dirname, 'fw/rssi-10040.hex'),
 *                 fwVersion: 'rssi-fw-1.0.0',
 *                 fwIdAddress: 0x2000,
 *             },
 *         },
 *         needSerialport: true,
 *         detailedOutput: false,
 *
 *         // called if programming is needed to be confirmed
 *         promiseConfirm: async message => (await inquirer.prompt([{
 *             type: 'confirm', name: 'isConfirmed', message, default: false,
 *         }])).isConfirmed,
 *
 *         // called if user need make a choice e.g. multiple DFU firmwares are defined
 *         promiseChoice: async (message, choices) => (await inquirer.prompt([{
 *             type: 'list', name: 'choice', message, choices,
 *         }])).choice,
 *     },
 * );
 *
 * @param {object} selectedDevice nrf-device-lister's device
 * @param {object} options Setup options, as shown in the example above. May include
 * { jprog, dfu, needSerialport, detailedOutput, promiseChoice, promiseConfirm }
 * @returns {Promise} Resolves with the device object that was set up, or
 * device and details if detailedOutput is enabled.
 */
function setupDevice(selectedDevice, options) {
    var jprog = options.jprog;
    var dfu = options.dfu;
    var needSerialport = options.needSerialport;
    var detailedOutput = options.detailedOutput;
    var promiseConfirm = options.promiseConfirm;

    if (dfu && Object.keys(dfu).length !== 0) {
        // check if device is in DFU-Bootlader, it might _only_ have serialport
        if (isDeviceInDFUBootloader(selectedDevice)) {
            debug$2('Device is in DFU-Bootloader, DFU is defined');
            return performDFU(selectedDevice, options);
        }

        var usbdevice = selectedDevice.usb;

        if (usbdevice) {
            var usbdev = usbdevice.device;
            var interfaceNumber = getDFUInterfaceNumber$1(usbdev);
            if (interfaceNumber >= 0) {
                debug$2('Device has DFU trigger interface, probably in Application mode');
                return getSemVersion$1(usbdev, interfaceNumber)
                    .then(function (semver) {
                        debug$2(("'" + semver + "'"));
                        if (Object.keys(dfu).map(function (key) { return dfu[key].semver; }).includes(semver)) {
                            if (needSerialport && !selectedDevice.serialport) {
                                return Promise.reject(new Error('Missing serial port'));
                            }
                            debug$2('Device is running the correct fw version');
                            return createReturnValue(
                                selectedDevice,
                                { wasProgrammed: false }, detailedOutput
                            );
                        }
                        debug$2('Device requires different firmware');
                        return performDFU(selectedDevice, options);
                    });
            }
            debug$2('Device is not in DFU-Bootloader and has no DFU trigger interface');
        }
    }


    if (jprog && selectedDevice.traits.includes('jlink')) {
        var wasProgrammed = false;
        return Promise.resolve()
            .then(function () { return needSerialport && verifySerialPortAvailable$1(selectedDevice); })
            .then(function () { return openJLink$1(selectedDevice); })
            .then(function () { return getDeviceInfo$1(selectedDevice); })
            .then(function (deviceInfo) {
                Object.assign(selectedDevice, { deviceInfo: deviceInfo });

                var family = (deviceInfo.family || '').toLowerCase();
                var deviceType = (deviceInfo.deviceType || '').toLowerCase();
                var shortDeviceType = deviceType.split('_').shift();
                var boardVersion = (selectedDevice.boardVersion || '').toLowerCase();

                var key = Object.keys(jprog).find(function (k) { return k.toLowerCase() === deviceType; })
                    || Object.keys(jprog).find(function (k) { return k.toLowerCase() === shortDeviceType; })
                    || Object.keys(jprog).find(function (k) { return k.toLowerCase() === boardVersion; })
                    || Object.keys(jprog).find(function (k) { return k.toLowerCase() === family; });

                if (!key) {
                    throw new Error('No firmware defined for selected device');
                }
                debug$2('Found matching firmware definition', key);
                return jprog[key];
            })
            .then(function (firmwareDefinition) { return __async(function*(){ return (
                {
                    valid: yield validateFirmware$1(selectedDevice, firmwareDefinition),
                    firmwareDefinition: firmwareDefinition,
                }
            )}()); })
            .then(function (ref) {
                var valid = ref.valid;
                var firmwareDefinition = ref.firmwareDefinition;

                if (valid) {
                    debug$2('Application firmware id matches');
                    return selectedDevice;
                }
                return confirmHelper(promiseConfirm)
                    .then(function (isConfirmed) {
                        if (!isConfirmed) {
                            // go on without update
                            return selectedDevice;
                        }
                        return programFirmware$1(selectedDevice, firmwareDefinition)
                            .then(function () {
                                wasProgrammed = true;
                            });
                    });
            })
            .then(
                function () { return closeJLink$1(selectedDevice).then(function () { return selectedDevice; }); },
                function (err) { return closeJLink$1(selectedDevice).then(function () { return Promise.reject(err); }); }
            )
            .then(function () { return createReturnValue(selectedDevice, { wasProgrammed: wasProgrammed }, detailedOutput); });
    }

    debug$2('Selected device cannot be prepared, maybe the app still can use it');
    return Promise.resolve(createReturnValue(
        selectedDevice,
        { wasProgrammed: false },
        detailedOutput
    ));
}

/* Copyright (c) 2010 - 2018, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

var index = {
    setupDevice: setupDevice,
    waitForDevice: waitForDevice,
    detachAndWaitFor: detachAndWaitFor,
    ensureBootloaderMode: ensureBootloaderMode,
    dfuTrigger: dfuTrigger,
    initPacket: initPacket,
};

module.exports = index;
//# sourceMappingURL=index.js.map
